package model.core;

import java.security.*;
import java.util.*;
import java.io.*;
import javax.persistence.*;
import org.hibernate.*;
import java.sql.*;

import model.util.*;


@Entity
@Table(name = "Users")
@SequenceGenerator(name = "seqId", sequenceName = "seqUserId")

@NamedQueries({

    @NamedQuery(
        name = "findByEmailPassword", 
        query = "SELECT user FROM User user WHERE user.email = :email AND user.password = :password"
    ), 

    @NamedQuery(
        name = "getLastProducts",
        query = "SELECT hasSeen.product FROM User user JOIN user.hasSeen hasSeen WHERE user.id = :userId ORDER BY hasSeen.timestamp DESC"
    )

})

public class User implements Serializable { 

    @Id
    @Column(name = "userId", nullable = false, unique = true)
    @GeneratedValue(strategy = GenerationType.AUTO, generator = "seqId")
    private Integer id;

    @Column
    private String firstName;

    @Column
    private String surname;
    
    @Column(unique = true, nullable = false)
    private String email;
    
    @Column(nullable = false)
    private String password;
    
    @Column
    private Boolean admin;

    @OneToMany(fetch = FetchType.EAGER)
    @JoinColumn(name = "userId")
    private Set<HasSeen> hasSeen = new HashSet();

    @OneToMany(mappedBy = "user", fetch = FetchType.EAGER)
    private Set<Purchase> purchases = new HashSet();


    public User() { }

    
    public Integer getId() { return this.id; }
    private void setId(Integer id) { this.id = id; }


    public String getFirstName() { return this.firstName; }
    public void setFirstName(String firstName) { this.firstName = firstName; }


    public String getSurname() { return this.surname; }
    public void setSurname(String surname) { this.surname = surname; }


    public String getEmail() { return this.email; }
    public void setEmail(String email) { this.email = email; }


    public String getPassword() { return this.password;}
    public void setPassword(String password) {

        try {

            // Get a message digestor implementing the SHA-1 algorithm 
            MessageDigest sha = MessageDigest.getInstance("SHA");

            // Make a hash of password and store it as a String
            this.password = new String(sha.digest(password.getBytes()), "UTF-8");

        } catch (Exception exception) {
            // Ignore exceptions as they will never happen
        }

    }


    public Boolean isAdmin() { return this.admin; }
    public void setAdmin(Boolean admin) { this.admin = admin; }


    public Set<HasSeen> getHasSeen() { return this.hasSeen; }
    private void setHasSeen(Set<HasSeen> hasSeen) {
        this.hasSeen = hasSeen;
    }

    public Set<Purchase> getPurchases() { return this.purchases; }
    private void setPurchases(Set<Purchase> purchases) {
        this.purchases = purchases;
    }


    public void add(Purchase purchase) {
       
        // Add to object
        this.purchases.add(purchase);
       
        // Associate
        if (purchase.getUser() != this) {
            purchase.setUser(this);
        }

        // Save in databas
		model.manager.UserManager.update(this);
        model.manager.PurchaseManager.update(purchase);

    }


    /**
     * State that an user has seen a product, creating an association that 
     * contains timestamp from user to product.
     *
     * @param product Persisted object representing the product the user has
     *                seen.
     */
    public void hasSeen(Product product) {

        // Create association
        HasSeen association = new HasSeen();
        association.setTimestamp(new Long(Calendar.getInstance().getTime().getTime()));
        association.setProduct(product);
        
        // Save association
        model.manager.HasSeenManager.add(association);

        // Associate user to this association
        this.hasSeen.add(association);
        
        // Update user in database
        model.manager.UserManager.update(this);

    }


    /**
     * Retrieve the last "maxResults" products this user has seen.
     *
     * @param maxResults Number of results to be retrieved.
     * @return An array of products containg last seen products.
     */
    public ArrayList<Product> getLastProducts(int maxResults) {
            
        ArrayList<Product> lastProducts = new ArrayList();

        // Get database connection
        Session session = HibernateUtil.getSession();
        session.beginTransaction();

        // Load query as specified in annotations by its name
        org.hibernate.Query query = session.getNamedQuery("getLastProducts");

        // Prepare query with parameters
        query.setInteger("userId", this.id);

        // Limit retrieval to only last maxResults results
        query.setMaxResults(maxResults);
        
        // For each retrieved result
        for (Iterator iterator = query.iterate(); iterator.hasNext(); ) {
            
            // Add to array
            lastProducts.add((Product) iterator.next());

        }

        // Commit transaction
        session.getTransaction().commit();

        return lastProducts;

    }

    

    private static void unitTest1() {
        
        User user = new User();
        user.setFirstName("firstName");
        user.setSurname("surname");
        user.setEmail("email");
        user.setPassword("password");
        user.setAdmin(false);
        
        model.manager.UserManager.add(user);

        Product product1 = new Product();
        product1.setTitle("teste1");
        product1.setDescription("teste1");

        Product product2 = new Product();
        product2.setTitle("teste2");
        
        model.manager.ProductManager.add(product1);
        model.manager.ProductManager.add(product2);

        user.hasSeen(product1);
        user.hasSeen(product2);

        Session session = HibernateUtil.getSession();

        session.beginTransaction();
        System.out.println("HasSeen from user: " + 
                ((User) session.get(User.class, user.getId())).getHasSeen().size());

        session.getTransaction().commit();
    
    }

    private static void unitTest2() {
    
        User user = new User();
        user.setFirstName("firstName");
        user.setSurname("surname");
        user.setEmail("email");
        user.setPassword("password");
        user.setAdmin(false);
        
        model.manager.UserManager.add(user);

        Product product1 = new Product();
        product1.setTitle("teste1");
        Product product2 = new Product();
        product2.setTitle("teste2");
        Product product3 = new Product();
        product3.setTitle("teste3");
        Product product4 = new Product();
        product4.setTitle("teste4");
        Product product5 = new Product();
        product5.setTitle("teste5");
        
        model.manager.ProductManager.add(product1);
        model.manager.ProductManager.add(product2);
        model.manager.ProductManager.add(product3);
        model.manager.ProductManager.add(product4);
        model.manager.ProductManager.add(product5);
        model.manager.ProductManager.add(product5);

        user.hasSeen(product5);
        user.hasSeen(product1);
        user.hasSeen(product2);
        user.hasSeen(product4);
        user.hasSeen(product3);

        model.manager.UserManager.update(user);

        ArrayList<Product> lastProducts = user.getLastProducts(3);

        for (int i = 0; i < lastProducts.size(); i++) {
            
            Product product = (Product) lastProducts.get(i);
            System.out.println("Title :" +  product.getTitle());

        }

    }

    private static void unitTest3() {
 
        // USER
        User user = new User();
        user.setFirstName("firstName");
        user.setSurname("surname");
        user.setEmail("email");
        user.setPassword("password");
        user.setAdmin(false);

        model.manager.UserManager.add(user);
        
        // PURCHASE
        Purchase purchase = new Purchase();
        purchase.setDate(new Long(Calendar.getInstance().getTime().getTime()));
        purchase.setPaid(true);
        purchase.setPaymentType("International credit card");
        
        model.manager.PurchaseManager.add(purchase);

        // ASSOCIATION
        user.add(purchase);
        
        System.out.println("User from purchase: " + 
                model.manager.PurchaseManager.findById(purchase.getId()).getUser().getFirstName());

        Session session = HibernateUtil.getSession();

        session.beginTransaction();
        System.out.println("Purchases from user: " + 
                ((User) session.get(User.class, user.getId())).getPurchases().size());

        session.getTransaction().commit();
    }

    public static void main(String args[]) {
        
        User.unitTest2();
    }

}

